#!/usr/bin/env python

'''
'''

__docformat__ = 'restructuredtext'
__version__ = '$Id: $'

import ctypes
import math

from pyglet import clock
from pyglet import font
from pyglet.gl import *
from pyglet.window import key
from pyglet.window import mouse

import collide
import euclid
import game
import menu
import obj
import threed

class SelectionHitRecord(ctypes.Structure):
    _fields_ = [
        ('stack_depth', GLuint),
        ('z1', GLuint),
        ('z2', GLuint),
        ('name', GLuint),  # Assume only one name (don't use stack)
    ]

class Character(object):
    MOVEMENT_NAMES = [
        'turn_left',
        'turn_right',
        'forward', 
        'left', 
        'backward', 
        'right', 
        'jump', 
        'fire']

    center_height = 0
    eye_height = 0
    step_height = 0
    turn_speed = 0
    move_speed = 0
    jump_speed = 0
    push_friction = 0
    collide_radius = 0
    collide_height = 0
    debug_color = (.7, .7, .7, 1)

    def __init__(self, position, heading, pitch):
        self.collision = \
            collide.Ellipsoid(position.copy(), 
                              euclid.Vector3(self.collide_radius, 
                                             self.collide_height, 
                                             self.collide_radius))
        self.heading = heading
        self.pitch = pitch

        self.velocity = euclid.Vector3(0, 0, 0)
        self.freefall = True
        self.pushed = False
        self.movement = {}
        for k in self.MOVEMENT_NAMES:
            self.movement[k] = False

    def get_facing_ray(self):
        return euclid.Ray3(self.collision.c,
            euclid.Quaternion.new_rotate_euler(
                math.radians(self.heading),
                0., 
                math.radians(self.pitch)) * euclid.Vector3(0, 0, -1))

down = euclid.Vector3(0, -1, 0)

class FPS(threed.ThreeD):
    MOUSE_SENSITIVITY = 0.6
    FOV = 90.
    KEYMAP = {
        key.W: 'forward',
        key.A: 'left',
        key.S: 'backward',
        key.D: 'right',
        key.SPACE: 'jump',
    }

    BUTTONMAP = {
        mouse.MOUSE_LEFT_BUTTON: 'fire',
    }

    def __init__(self, level):
        self.level = level
        self.restart()

        self.collision = collide.Mesh(level.collision_meshes)

        self.debug_view_collision = False       # Enable debug visualisation
        self.debug_colliding = False            # Is currently hit something?
        self.debug_quadric = gluNewQuadric()

        self.clock_display = None

        game_font = font.load('Arial', 36, bold=True)
        self.game_label = font.Label(game_font, '', color=(.9, .9, 1, .9))

    def restart(self):
        self.level.restart()
        self.player = self.level.player # convenience

    def activate(self, window):
        super(FPS, self).activate(window)

        glClearAccum(0, 0, 0, 0)
        glClear(GL_ACCUM_BUFFER_BIT)

        self.beam_opacity = 0.

        for k in self.player.movement.keys():
            self.player.movement[k] = False

    def update(self, dt):
        self.debug_colliding = False

        # AI, game state
        self.level.update(dt)
        if self.level.has_exit:
            self.has_exit = True
            return

        # Beam fades out
        self.beam_opacity = max(0, self.beam_opacity - dt * 2)

        # Update particle emitters
        for emitter in self.level.emitters:
            emitter.update(dt)

        # Everybody moves, collides with environment
        self.update_character(dt, self.player)
        for baddie in self.level.baddies:
            self.update_character(dt, baddie)

        # Player collides with baddies
        for baddie in self.level.baddies:
            if self.player.collision.contact_ellipsoid(baddie.collision):
                self.debug_colliding = True
                self.level.on_collide_baddie(baddie)

        # Player collides with triggers
        for polygon, name in self.level.triggers:
            if polygon.contact_ellipsoid(self.player.collision):
                self.level.on_trigger(dt, name)

        # Trigger cutscene
        if self.level.queued_state:
            self.queued_state = self.level.queued_state
            self.level.queued_state = None

        # Trigger game over
        if self.level.dead:
            self.queued_state = menu.GameMenu(self)

    def update_character(self, dt, character):
        p = character

        # Constant-rate turning (baddies, or player keyboard control)
        dr = (p.movement['turn_right'] - p.movement['turn_left'])
        p.heading += dr * dt * p.turn_speed

        # Horizontal movement
        dz = (p.movement['backward'] - p.movement['forward'])
        dx = (p.movement['right'] - p.movement['left'])
        # [ cos -sin ] [dx] = [dx']
        # [ sin  cos ] [dz]   [dz']
        cos = math.cos(-math.radians(p.heading))
        sin = math.sin(-math.radians(p.heading))
        dx, dz = dx * cos - dz * sin, dx * sin + dz * cos
        if p.freefall:
            dx /= 2
            dz /= 2

        # Move the collision ellipsoid now, look for collisions later.
        if not p.pushed:
            p.collision.c.x += dx * dt * p.move_speed
            p.collision.c.z += dz * dt * p.move_speed

        # Freefall movement due to gravity
        if p.freefall or p.pushed:
            p.velocity += dt * self.level.gravity
            p.collision.c += dt * p.velocity
            if p.collision.c.y < self.level.fallout_height:
                self.level.on_character_fallout(p)
                return

        # See if the character walked into a wall
        resolution = euclid.Vector3(0, 0, 0)
        collisions = self.collision.collide_ellipsoid(p.collision)
        for plane, d in collisions:
            self.debug_colliding = True
            r = d * plane.n
            if plane.n.dot(down) > .5:
                # Hit head on roof
                r.y = min(0, r.y) # Don't apply floor collisions
                p.velocity.y = 0
            resolution += r
        p.collision.c += resolution

        # If falling or moving, test the ground beneath character's feet
        if p.freefall or dx or dz:
            # Check for falling through the floor at high velocity
            seg_length = max(p.step_height, -dt * p.velocity.y)
            segment = euclid.LineSegment3(
                euclid.Point3(p.collision.c.x, 
                              p.collision.c.y - p.center_height, 
                              p.collision.c.z),
                euclid.Vector3(0, seg_length, 0))
            point, plane = self.collision.collide_floor(segment)
            if plane:
                p.collision.c.y = point.y + p.center_height
                p.velocity.y = 0
                p.freefall = False
            else:
                p.freefall = True

        # Remove velocity when on ground (so it doesn't accumulate)
        if not p.pushed and not p.freefall:
            p.velocity = euclid.Vector3(0, 0, 0)
        elif p.pushed and (not p.freefall or resolution):
            # If pushed, apply floor or wall friction
            p.velocity *= (1 - dt) * (1 - character.push_friction)
            if p.velocity.magnitude_squared() < 0.1:
                p.velocity = euclid.Vector3(0, 0, 0)
                p.pushed = False

        # Begin a jump if requested and possible
        if p.movement['jump'] and not p.freefall and not p.pushed:
            p.velocity.x = dx * p.move_speed
            p.velocity.y = p.jump_speed
            p.velocity.z = dz * p.move_speed
            p.freefall = True

    def draw(self):
        glClearColor(1, 1, 1, 1)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        self.draw_skybox(self.level.skybox, 
            self.player.pitch, self.player.heading)
        self.apply_camera()

        self.draw_inner()

        for emitter in self.level.emitters:
            emitter.draw()

        self.draw_weapon()

        if self.beam_opacity:
            self.draw_beam()

        if self.level.motion_blur:
            glAccum(GL_MULT, self.level.motion_blur)
            glAccum(GL_ACCUM, 1 - self.level.motion_blur)
            glAccum(GL_RETURN, 1.)

        # HUD
        self.begin2d()
        crosshair = self.level.crosshair
        glPushAttrib(GL_ENABLE_BIT)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        glEnable(crosshair.target)
        crosshair.blit((self._width  - crosshair.width) / 2,
                       (self._height - crosshair.height) / 2, 0)
        glPopAttrib()

        if self.level.game_text:
            self.game_label.text = self.level.game_text
            self.game_label.x = (self._width - self.game_label.width) / 2
            self.game_label.y = self._height / 3.5
            self.game_label.draw()

        if self.level.fade_to_black:
            self.draw_fade(self.level.fade_to_black)

        if self.clock_display:
            self.clock_display.draw()
        self.end2d()

        if self.player.movement['fire'] and self.level.on_fire():
            self.beam_opacity = 1.

            # Draw in select mode, see what was clicked
            select_buf = (SelectionHitRecord * 32)()
            select_buf_p = cast(select_buf, ctypes.POINTER(GLuint))
            glSelectBuffer(len(select_buf), select_buf_p)
            glRenderMode(GL_SELECT)

            viewport = (GLint * 4)()
            glGetIntegerv(GL_VIEWPORT, viewport)
            glMatrixMode(GL_PROJECTION)
            glPushMatrix()
            glLoadIdentity()
            gluPickMatrix(self._width / 2, self._height / 2, 
                2., 2., viewport)
            self.apply_projection()
            glMatrixMode(GL_MODELVIEW)

            glInitNames()
            glPushName(0)
            self.apply_camera()
            self.draw_inner(materials=False, name_baddies=True)
            glFlush()

            glMatrixMode(GL_PROJECTION)
            glPopMatrix()
            glMatrixMode(GL_MODELVIEW)

            hits = glRenderMode(GL_RENDER)
            if hits == -1:
                warnings.warn('Too many selection hits.')
            elif hits > 0:
                records = select_buf[:hits]
                records.sort(lambda a,b: cmp(a.z1, b.z1))
                if records[0].name > 0:
                    ray = self.player.get_facing_ray()
                    self.level.on_hit_baddie(
                        self.level.baddies[records[0].name - 1], ray)


    def apply_camera(self):
        glLoadIdentity()
        if self.debug_view_collision:
            glTranslatef(0, 0, -2)
        p = self.player
        glRotatef(-p.pitch, 1, 0, 0)
        glRotatef(-p.heading, 0, 1, 0)
        glTranslatef(-p.collision.c.x, 
                     -p.collision.c.y - p.eye_height,
                     -p.collision.c.z)
        
    def draw_inner(self, materials=True, name_baddies=False):
        glEnable(GL_DEPTH_TEST)

        if materials:
            glEnable(GL_LIGHTING)
            glEnable(GL_LIGHT0)
            glLightfv(GL_LIGHT0, GL_POSITION, (GLfloat * 4)(0, 1, 0, 0))
            glLightfv(GL_LIGHT0, GL_AMBIENT, (GLfloat * 4)(.2, .2, .2, 1))
            glLightfv(GL_LIGHT0, GL_DIFFUSE, (GLfloat * 4)(.5, .5, .5, 1))
            glLightfv(GL_LIGHT0, GL_SPECULAR, (GLfloat * 4)(0, 0, 0, 1))

            glEnable(GL_LIGHT1)
            glLightfv(GL_LIGHT1, GL_POSITION, (GLfloat * 4)(0, -1, .1, 0))
            glLightfv(GL_LIGHT1, GL_DIFFUSE, (GLfloat * 4)(.2, .2, .2, 1))
            glLightfv(GL_LIGHT0, GL_SPECULAR, (GLfloat * 4)(0, 0, 0, 1))
       
        # Visible geometry
        for mesh in self.level.visual_meshes:
            mesh.draw()

        # Baddies
        for i, baddie in enumerate(self.level.baddies):
            if name_baddies:
                glLoadName(i + 1)
            glPushMatrix()
            glTranslatef(*baddie.collision.c)
            glRotatef(baddie.heading, 0, 1, 0)
            baddie.draw()
            glPopMatrix()
        glLoadName(0)

        # Debug view
        if self.debug_view_collision:
            glPushAttrib(GL_LIGHTING_BIT | GL_ENABLE_BIT)

            # Debug view of step height
            if p.freefall:
                glMaterialfv(GL_FRONT, GL_DIFFUSE, (GLfloat * 4)(1, 0, 0, 1))
            else:
                glMaterialfv(GL_FRONT, GL_DIFFUSE, (GLfloat * 4)(0, 1, 0, 1))
            glPushMatrix()
            glTranslatef(p.collision.c.x, 
                         p.collision.c.y - p.center_height,
                         p.collision.c.z) 
            gluSphere(self.debug_quadric, p.step_height, 50, 50)
            glPopMatrix()

            # Debug view of collision ellipsoid
            glPushMatrix()
            glTranslatef(*p.collision.c)
            if self.debug_colliding:
                glMaterialfv(GL_FRONT, GL_DIFFUSE, (GLfloat * 4)(1, 0, 0, 1))
            else:
                glMaterialfv(GL_FRONT, GL_DIFFUSE, (GLfloat * 4)(0, 1, 0, 1))
            glScalef(*p.collision.v)
            glEnable(GL_NORMALIZE)
            gluSphere(self.debug_quadric, 1, 50, 50)
            glPopMatrix()
            glPopAttrib()

    def draw_weapon(self):
        glPushMatrix()
        glLoadIdentity()

        glTranslatef(.3, -.5, -.35)
        for mesh in self.level.weapon_meshes:
            mesh.draw()

        glPopMatrix()

    def draw_beam(self):
        beam = self.level.beam_texture
        glPushMatrix()
        glLoadIdentity()

        glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT)
        glDisable(GL_LIGHTING)

        glEnable(GL_BLEND)
        glEnable(beam.target)
        glBindTexture(beam.target, beam.id)

        glColor4f(1, 1, 1, self.beam_opacity)

        w = 0.05
        x = 0.3
        y = -.5
        z = -.35
        glBegin(GL_QUADS)
        glTexCoord2f(0, 0)
        glVertex3f(x-w, y, z)
        glTexCoord2f(1, 0)
        glVertex3f(x+w, y, z)
        glTexCoord2f(1, 1000)
        glVertex3f(x+w, 0, -1000)
        glTexCoord2f(0, 1000)
        glVertex3f(x-w, 0, -1000)
        glEnd()
        glPopAttrib()

        glPopMatrix()

    def on_mouse_motion(self, x, y, dx, dy):
        self.player.pitch += dy * self.MOUSE_SENSITIVITY
        self.player.heading -= dx * self.MOUSE_SENSITIVITY

        self.player.pitch = min(max(self.player.pitch, -80), 80)

    def on_mouse_drag(self, x, y, dx, dy, button, modifiers):
        self.on_mouse_motion(x, y, dx, dy)

    def on_mouse_press(self, button, x, y, modifiers):
        if button in self.BUTTONMAP:
            self.player.movement[self.BUTTONMAP[button]] = True

    def on_mouse_release(self, button, x, y, modifiers):
        if button in self.BUTTONMAP:
            self.player.movement[self.BUTTONMAP[button]] = False

    def on_key_press(self, symbol, modifiers):
        if symbol in self.KEYMAP:
            self.player.movement[self.KEYMAP[symbol]] = True
        elif symbol == key.ESCAPE:
            self.queued_state = menu.GameMenu(self)
        elif symbol == key.T and \
             key.MOD_CTRL & modifiers and \
             key.MOD_SHIFT & modifiers and \
             self.level.allow_cheat:
            self.queued_state = menu.CheatMenu(self)

        return True

    def on_key_release(self, symbol, modifiers):
        if symbol in self.KEYMAP:
            self.player.movement[self.KEYMAP[symbol]] = False
