#!/usr/bin/env python

'''
'''

__docformat__ = 'restructuredtext'
__version__ = '$Id: $'

import random

from pyglet.gl import *

import game

class ThreeD(game.GameState):
    FOV = 90.
    exclusive_mouse = True

    def activate(self, window):
        if self.exclusive_mouse:
            window.set_exclusive_mouse()

        self._width = window.width
        self._height = window.height
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        self.apply_projection()
        glMatrixMode(GL_MODELVIEW)

    def deactivate(self, window):
        if self.exclusive_mouse:
            window.set_exclusive_mouse(False)

    def apply_projection(self):
        gluPerspective(self.FOV, self._width / float(self._height), .1, 1000)

    def begin2d(self):
        glPushAttrib(GL_ENABLE_BIT)
        glDisable(GL_DEPTH_TEST)
        glDisable(GL_LIGHTING)

        glMatrixMode(GL_PROJECTION)
        glPushMatrix()
        glLoadIdentity()
        glOrtho(0, self._width, 0, self._height, -1, 1)

        glMatrixMode(GL_MODELVIEW)
        glPushMatrix()
        glLoadIdentity()

    def end2d(self):
        glPopMatrix()
        glMatrixMode(GL_PROJECTION)

        glPopMatrix()
        glMatrixMode(GL_MODELVIEW)

        glPopAttrib()

    def draw_skybox(self, textures, pitch, heading):
        # Camera
        glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT)
        glDisable(GL_DEPTH_TEST)
        glDisable(GL_LIGHTING)
        glColor3f(1, 1, 1)

        glPushMatrix()
        glLoadIdentity()
        glRotatef(-pitch, 1, 0, 0)
        glRotatef(-heading, 0, 1, 0)

        r = 1. # r for readability
        coords = [
            (( r, -r, -1), ( r,  r, -1), (-r,  r, -1), (-r, -r, -1)), # ft
            ((-r, -r,  1), (-r,  r,  1), ( r,  r,  1), ( r, -r,  1)), # bk
            ((-1, -r, -r), (-1,  r, -r), (-1,  r,  r), (-1, -r,  r)), # lt
            (( 1, -r,  r), ( 1,  r,  r), ( 1,  r, -r), ( 1, -r, -r)), # rt
            (( r,  1,  r), (-r,  1,  r), (-r,  1, -r), ( r,  1, -r)), # up
            ((-r, -1, -r), (-r, -1,  r), ( r, -1,  r), ( r, -1, -r)), # dn
        ]

        for texture, (p1, p2, p3, p4) in map(None, textures, coords):
            glEnable(texture.target)
            glBindTexture(texture.target, texture.id)
            glTexParameteri(texture.target, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE)
            glTexParameteri(texture.target, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE)
            glBegin(GL_QUADS)
            glTexCoord2f(0., 0.)
            glVertex3f(*p1)
            glTexCoord2f(0., 1.)
            glVertex3f(*p2)
            glTexCoord2f(1., 1.)
            glVertex3f(*p3)
            glTexCoord2f(1., 0.)
            glVertex3f(*p4)
            glEnd()

        glPopMatrix()
        glPopAttrib()

    def draw_fade(self, alpha):
        glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT)
        glEnable(GL_BLEND)
        glDisable(GL_TEXTURE_2D)
        glColor4f(0, 0, 0, alpha)
        glBegin(GL_QUADS)
        glVertex2f(0, 0)
        glVertex2f(self._width, 0)
        glVertex2f(self._width, self._height)
        glVertex2f(0, self._height)
        glEnd()
        glPopAttrib()

class Particle(object):
    __slots__ = ['position', 'velocity']

    def __init__(self, position, velocity):
        self.position = position.copy()
        self.velocity = velocity.copy()
 
class Emitter(object):
    def __init__(self, 
                 position, 
                 position_sigma, 
                 velocity, 
                 velocity_sigma, 
                 texture, 
                 point_size,
                 min_size,
                 period, 
                 max_particles):
        self.position = position
        self.position_sigma = position_sigma
        self.velocity = velocity
        self.velocity_sigma = velocity_sigma
        self.texture = texture
        self.point_size = point_size
        self.min_size = min_size
        self.period = period
        self.max_particles = max_particles

        self.particles = []
        self.time = 0.

    def update(self, dt):
        self.time += dt
        if self.time > self.period:
            # Add a particle
            self.time -= self.period
            p = self.position.copy()
            if self.position_sigma:
                p.x += random.gauss(0, self.position_sigma.x)
                p.y += random.gauss(0, self.position_sigma.y)
                p.z += random.gauss(0, self.position_sigma.z)
            v = self.velocity.copy()
            if self.velocity_sigma:
                v.x += random.gauss(0, self.velocity_sigma.x)
                v.y += random.gauss(0, self.velocity_sigma.y)
                v.z += random.gauss(0, self.velocity_sigma.z)
            self.particles.append(Particle(p, v))

            # Remove a particle
            if len(self.particles) > self.max_particles:
                self.particles.pop(0)

        for particle in self.particles:
            particle.position += particle.velocity * dt

    def draw(self):
        glDepthMask(False)

        glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT)
        glColor3f(1, 1, 1)
        glDisable(GL_LIGHTING)


        glEnable(GL_BLEND)
        glEnable(self.texture.target)

        point_size = GLfloat()
        glGetFloatv(GL_POINT_SIZE_MAX_ARB, point_size)
        glBindTexture(self.texture.target, self.texture.id)
        glPointSize(self.point_size)
        glPointParameterfvARB(GL_POINT_DISTANCE_ATTENUATION_ARB,
            (GLfloat * 3)(0, 0, .01))
        glPointParameterfARB(GL_POINT_SIZE_MIN_ARB, self.min_size)
        glTexEnvf(GL_POINT_SPRITE_ARB, GL_COORD_REPLACE_ARB, GL_TRUE)
        glEnable(GL_POINT_SPRITE_ARB)

        array = []
        for p in self.particles:
            array += [p.position.x, p.position.y, p.position.z]
        self.array = (GLfloat * len(array))(*array)

        glPushClientAttrib(GL_CLIENT_VERTEX_ARRAY_BIT)
        glVertexPointer(3, GL_FLOAT, 0, self.array)
        glEnableClientState(GL_VERTEX_ARRAY)
        glDrawArrays(GL_POINTS, 0, len(self.particles))
        glPopClientAttrib()
        
        glPopAttrib()

        glDepthMask(True)
